@@ -21,6 +21,7 @@ from cloudvolume.lib import Vec, Bbox, max2, min2, xyzrange, find_closest_diviso
 from cloudvolume.datasource.precomputed.sharding import ShardingSpecification
 from cloudfiles import CloudFiles
 import cloudfiles.paths
+import shardcomputer
 
 from igneous.tasks import (
   MeshTask, MeshManifestPrefixTask, 
@@ -609,12 +610,8 @@ def create_sharded_multires_mesh_from_unsharded_tasks(
 
   cv_dest = CloudVolume(dest, mesh_dir=mesh_dir)
 
-  # perf: ~66.5k hashes/sec on M1 ARM64
-  shardfn = lambda lbl: cv_dest.mesh.reader.spec.compute_shard_location(lbl).shard_number
-
-  shard_labels = defaultdict(list)
-  for label in tqdm(all_labels, desc="Hashes"):
-    shard_labels[shardfn(label)].append(label)
+  all_labels = np.fromiter(all_labels, dtype=np.uint64, count=len(all_labels))
+  shard_labels = shardcomputer.assign_labels_to_shards(all_labels, preshift_bits, shard_bits, minishard_bits)
   del all_labels
 
   cf = CloudFiles(cv_dest.mesh.meta.layerpath, progress=True)
@@ -668,7 +665,8 @@ def create_sharded_multires_mesh_tasks(
   minishard_index_encoding="gzip", 
   mesh_dir:Optional[str] = None, 
   spatial_index_db:Optional[str] = None,
-  min_chunk_size:Tuple[int,int,int] = (512,512,512)
+  min_chunk_size:Tuple[int,int,int] = (512,512,512),
+  max_labels_per_shard:Optional[int] = None,
 ) -> Iterator[MultiResShardedMeshMergeTask]: 
 
   configure_multires_info(
@@ -684,6 +682,10 @@ def create_sharded_multires_mesh_tasks(
   # 17 sec to download for pinky100
   all_labels = cv.mesh.spatial_index.query(cv.bounds * cv.resolution)
   
+  if max_labels_per_shard is not None:
+    assert max_labels_per_shard >= 1
+    min_shards = max(int(np.ceil(len(all_labels) / max_labels_per_shard)), min_shards)
+
   (shard_bits, minishard_bits, preshift_bits) = \
     compute_shard_params_for_hashed(
       num_labels=len(all_labels),
@@ -707,12 +709,8 @@ def create_sharded_multires_mesh_tasks(
 
   cv = CloudVolume(cloudpath)
 
-  # perf: ~66.5k hashes/sec on M1 ARM64
-  shardfn = lambda lbl: cv.mesh.reader.spec.compute_shard_location(lbl).shard_number
-
-  shard_labels = defaultdict(list)
-  for label in tqdm(all_labels, desc="Hashes"):
-    shard_labels[shardfn(label)].append(label)
+  all_labels = np.fromiter(all_labels, dtype=np.uint64, count=len(all_labels))
+  shard_labels = shardcomputer.assign_labels_to_shards(all_labels, preshift_bits, shard_bits, minishard_bits)
   del all_labels
 
   cf = CloudFiles(cv.mesh.meta.layerpath, progress=True)
