@@ -1,6 +1,7 @@
 from typing import Optional, List, Dict, Tuple
 
 from collections import defaultdict
+from concurrent.futures import ThreadPoolExecutor
 import functools
 import itertools
 import json
@@ -14,7 +15,7 @@ import struct
 import numpy as np
 from tqdm import tqdm
 
-from cloudfiles import CloudFiles
+from cloudfiles import CloudFiles, CloudFile
 
 from cloudvolume import CloudVolume, Mesh, view
 from cloudvolume.lib import Vec, Bbox, jsonify, sip, toiter, first
@@ -149,14 +150,18 @@ def process_mesh(
 
       # mesh.vertices must be integer type or mesh will display
       # distorted in neuroglancer.
-      submesh = DracoPy.encode(
-        submesh.vertices, submesh.faces, 
-        quantization_bits=vqb,
-        compression_level=draco_compression_level,
-        quantization_range=quantization_range,
-        quantization_origin=minpt,
-        create_metadata=True,
-      )
+      try:
+        submesh = DracoPy.encode(
+          submesh.vertices, submesh.faces, 
+          quantization_bits=vqb,
+          compression_level=draco_compression_level,
+          quantization_range=quantization_range,
+          quantization_origin=minpt,
+          create_metadata=True,
+        )
+      except DracoPy.EncodingFailedException:
+        continue
+
       manifest.fragment_offsets.append(len(submesh))
       mesh_binaries.append(submesh)
 
@@ -378,7 +383,7 @@ def collect_mesh_fragments(
   dirfn = lambda loc: cv.meta.join(mesh_dir, loc)
   filenames = [ dirfn(loc) for loc in filenames ]
 
-  block_size = 50
+  block_size = 20
 
   if len(filenames) < block_size:
     blocks = [ filenames ]
@@ -388,29 +393,41 @@ def collect_mesh_fragments(
     blocks = sip(filenames, block_size)
 
   all_meshes = defaultdict(list)
+  def process_shardfile(item):
+    filename, content = item
+    fragment = MapBuffer(content, frombytesfn=Mesh.from_precomputed)
+    fragment.validate()
+
+    for label in labels:
+      try:
+        mesh = fragment[label]
+        mesh.id = label
+        all_meshes[label].append((filename, mesh))
+      except KeyError:
+        continue
+
+    if hasattr(content, "close"):
+      content.close()
+
+    return filename
+
   for filenames_block in tqdm(blocks, desc="Filename Block", total=n_blocks, disable=(not progress)):
     if cv.meta.path.protocol == "file":
       all_files = {}
       prefix = cv.cloudpath.replace("file://", "")
       for filename in filenames_block:
         all_files[filename] = open(os.path.join(prefix, filename), "rb")
+
+      for item in tqdm(all_files.items(), desc="Scanning Fragments", disable=(not progress)):
+        process_shardfile(item)
     else:
-      all_files = cv.mesh.cache.download(filenames_block, progress=progress)
-    
-    for filename, content in tqdm(all_files.items(), desc="Scanning Fragments", disable=(not progress)):
-      fragment = MapBuffer(content, frombytesfn=Mesh.from_precomputed)
-      fragment.validate()
-
-      for label in labels:
-        try:
-          mesh = fragment[label]
-          mesh.id = label
-          all_meshes[label].append((filename, mesh))
-        except KeyError:
-          continue
-
-      if hasattr(content, "close"):
-        content.close()
+      all_files = { 
+        filename: CloudFile(cv.meta.join(cv.cloudpath, filename), cache_meta=True) 
+        for filename in filenames_block 
+      } 
+      with ThreadPoolExecutor(max_workers=block_size) as executor:
+        for filename in executor.map(process_shardfile, all_files.items()):
+          pass
 
   # ensure consistent results across multiple runs
   # by sorting mesh fragments by filename
@@ -452,7 +469,7 @@ def labels_for_shard(
     if spec.compute_shard_location(lbl).shard_number == shard_no 
   ]
   
-## Below functons adapted from 
+## Below functions adapted from 
 ## https://github.com/google/neuroglancer/issues/272
 ## Thanks to Hythem Sidky (@hsidky) for sharing his 
 ## progress and code with the connectomics community.
