@@ -57,9 +57,14 @@ class Tuple2(click.ParamType):
         self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
     return value
 
+class CloudPath(click.ParamType):
+  name = "CloudPath"
+  def convert(self, value, param, ctx):
+    return cloudfiles.paths.normalize(value)
+
 @click.group()
 @click.option("-p", "--parallel", default=1, help="Run with this number of parallel processes. If 0, use number of cores.")
-@click.version_option(version="4.6.0")
+@click.version_option(version="4.10.0")
 @click.pass_context
 def main(ctx, parallel):
   """
@@ -109,7 +114,7 @@ def imagegroup():
   pass
 
 @imagegroup.command()
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
 @click.option('--mip', default=0, help="Build upward from this level of the image pyramid. Default: 0")
 @click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@@ -148,8 +153,6 @@ def downsample(
   current top mip level of the pyramid. This builds it even taller
   (referred to as "superdownsampling").
   """
-  path = cloudfiles.paths.normalize(path)
-
   if sharded and num_mips != 1:
     print("igneous: sharded downsamples only support producing one mip at a time.")
     return
@@ -199,8 +202,8 @@ def downsample(
   tq.insert(tasks, parallel=parallel)
 
 @imagegroup.command()
-@click.argument("src")
-@click.argument("dest")
+@click.argument("src", type=CloudPath())
+@click.argument("dest", type=CloudPath())
 @click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
 @click.option('--mip', default=0, help="Build upward from this level of the image pyramid.", show_default=True)
 @click.option('--translate', type=Tuple3(), default=(0, 0, 0), help="Translate the bounding box by X,Y,Z voxels in the new location.")
@@ -209,6 +212,7 @@ def downsample(
 @click.option('--memory', default=3.5e9, type=int, help="Task memory limit in bytes. Task shape will be chosen to fit and maximize downsamples.", show_default=True)
 @click.option('--max-mips', default=5, help="Maximum number of additional pyramid levels.", show_default=True)
 @click.option('--encoding', default="raw", help="Which image encoding to use. Options: [all] raw, png; [images] jpeg; [segmentations] cseg, compresso; [floats] fpzip, kempressed", show_default=True)
+@click.option('--encoding-level', default=None, help="For some encodings (png level,jpeg quality,fpzip precision) a simple scalar value can adjust the compression efficiency.", show_default=True)
 @click.option('--sparse', is_flag=True, default=False, help="Don't count black pixels in mode or average calculations. For images, eliminates edge ghosting in 2x2x2 downsample. For segmentation, prevents small objects from disappearing at high mip levels.")
 @click.option('--shape', type=Tuple3(), default=(2048, 2048, 64), help="(overrides --memory) Set the task shape in voxels. This also determines how many downsamples you get. e.g. 2048,2048,64")
 @click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of destination layer. e.g. 128,128,64")
@@ -220,14 +224,16 @@ def downsample(
 @click.option('--dest-voxel-offset', type=Tuple3(), default=None, help="Set the voxel offset for this mip level.")
 @click.option('--clean-info', is_flag=True, default=False, help="Scrub info file of mesh and skeleton fields.", show_default=True)
 @click.option('--no-src-update', is_flag=True, default=False, help="Don't update the source provenance file with the transfer metadata.", show_default=True)
+@click.option('--truncate-scales/--no-truncate-scales', is_flag=True, default=True, help="Truncate the info file scales to the mip specified.", show_default=True)
 @click.pass_context
 def xfer(
 	ctx, src, dest, queue, translate, 
   downsample, mip, fill_missing, 
   memory, max_mips, shape, sparse, 
-  encoding, chunk_size, compress, 
+  encoding, encoding_level, chunk_size, compress,
   volumetric, delete_bg, bg_color, sharded,
-  dest_voxel_offset, clean_info, no_src_update
+  dest_voxel_offset, clean_info, no_src_update,
+  truncate_scales
 ):
   """
   Copy, re-encode, or shard an image layer.
@@ -247,9 +253,6 @@ def xfer(
   Use the --memory flag to automatically compute the
   a reasonable task shape based on your memory limits.
   """
-  src = cloudfiles.paths.normalize(src)
-  dest = cloudfiles.paths.normalize(dest)
-
   if encoding == "cseg":
     encoding = "compressed_segmentation"
 
@@ -265,7 +268,8 @@ def xfer(
       src, dest,
       chunk_size=chunk_size, fill_missing=fill_missing, mip=mip, 
       dest_voxel_offset=dest_voxel_offset, translate=translate, 
-      encoding=encoding, memory_target=memory, clean_info=clean_info
+      encoding=encoding, memory_target=memory, clean_info=clean_info,
+      encoding_level=encoding_level, truncate_scales=truncate_scales,
     )
   else:
     tasks = tc.create_transfer_tasks(
@@ -276,7 +280,8 @@ def xfer(
       delete_black_uploads=delete_bg, background_color=bg_color,
       compress=compress, factor=factor, sparse=sparse,
       memory_target=memory, max_mips=max_mips, 
-      clean_info=clean_info, no_src_update=no_src_update
+      clean_info=clean_info, no_src_update=no_src_update,
+      encoding_level=encoding_level, truncate_scales=truncate_scales,
     )
 
   parallel = int(ctx.obj.get("parallel", 1))
@@ -289,7 +294,7 @@ def contrastgroup():
   pass
 
 @contrastgroup.command()
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
 @click.option('--mip', default=0, help="Build histogram from this level of the image pyramid.", show_default=True)
 @click.option('--coverage', default=0.01, type=float, help="Fraction of the image to sample. Range: [0,1]", show_default=True)
@@ -303,8 +308,6 @@ def histogram(
   xrange, yrange, zrange, bounds_mip
 ):
   """(1) Compute the histogram for each z-slice."""
-  path = cloudfiles.paths.normalize(path)
-
   bounds = None
   if xrange or yrange or zrange:
     bounds = CloudVolume(path).meta.bounds(mip)
@@ -333,8 +336,8 @@ def histogram(
   tq.insert(tasks, parallel=parallel)
 
 @contrastgroup.command()
-@click.argument("src")
-@click.argument("dest")
+@click.argument("src", type=CloudPath())
+@click.argument("dest", type=CloudPath())
 @click.option('--shape', default="2048,2048,64", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
 @click.option('--translate', type=Tuple3(), default=(0, 0, 0), help="Translate the bounding box by X,Y,Z voxels in the new location.")
 @click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@@ -356,9 +359,6 @@ def equalize(
   xrange, yrange, zrange, bounds_mip
 ):
   """(2) Apply histogram equalization to z-slices."""
-  src = cloudfiles.paths.normalize(src)
-  dest = cloudfiles.paths.normalize(dest)
-
   bounds = None
   if xrange or yrange or zrange:
     bounds = CloudVolume(path).meta.bounds(mip)
@@ -414,7 +414,7 @@ def cclgroup():
   pass
 
 @cclgroup.command("faces")
-@click.argument("src")
+@click.argument("src", type=CloudPath())
 @click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
 @click.option('--mip', default=0, help="Apply to this level of the image pyramid.", show_default=True)
 @click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@@ -429,7 +429,6 @@ def ccl_faces(
   fill_missing, dust
 ):
   """(1) Generate back face images."""
-  src = cloudfiles.paths.normalize(src)
   tasks = tc.create_ccl_face_tasks(
     src, mip, shape,
     threshold_lte=intify(threshold_lte),
@@ -443,7 +442,7 @@ def ccl_faces(
   tq.insert(tasks, parallel=parallel)
 
 @cclgroup.command("links")
-@click.argument("src")
+@click.argument("src", type=CloudPath())
 @click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
 @click.option('--mip', default=0, help="Apply to this level of the image pyramid.", show_default=True)
 @click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@@ -458,7 +457,6 @@ def ccl_equivalences(
   fill_missing, dust
 ):
   """(2) Generate links between tasks."""
-  src = cloudfiles.paths.normalize(src)
   tasks = tc.create_ccl_equivalence_tasks(
     src, mip, shape,
     threshold_lte=intify(threshold_lte),
@@ -472,19 +470,18 @@ def ccl_equivalences(
   tq.insert(tasks, parallel=parallel)
 
 @cclgroup.command("calc-labels")
-@click.argument("src")
+@click.argument("src", type=CloudPath())
 @click.option('--mip', default=0, required=True, help="Apply to this level of the image pyramid.", show_default=True)
 @click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
 @click.pass_context
 def ccl_calc_labels(ctx, src, mip, shape):
   """(3) Compute and serialize a relabeling to the DB."""
   import igneous.tasks.image.ccl
-  src = cloudfiles.paths.normalize(src)
   igneous.tasks.image.ccl.create_relabeling(src, mip, shape)
 
 @cclgroup.command("relabel")
-@click.argument("src")
-@click.argument("dest")
+@click.argument("src", type=CloudPath())
+@click.argument("dest", type=CloudPath())
 @click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
 @click.option('--mip', default=0, help="Apply to this level of the image pyramid.", show_default=True)
 @click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of destination layer. e.g. 128,128,64")
@@ -503,8 +500,6 @@ def ccl_relabel(
   fill_missing, dust
 ):
   """(4) Finally relabel and write a CCL image."""
-  src = cloudfiles.paths.normalize(src)
-  dest = cloudfiles.paths.normalize(dest)
   tasks = tc.create_ccl_relabel_tasks(
     src, dest, 
     mip=mip, shape=shape, 
@@ -520,17 +515,16 @@ def ccl_relabel(
   tq.insert(tasks, parallel=parallel)
 
 @cclgroup.command("clean")
-@click.argument("src")
+@click.argument("src", type=CloudPath())
 @click.option('--mip', default=0, required=True, help="Apply to this level of the image pyramid.", show_default=True)
 def ccl_clean(src, mip):
   """(last) Cleans up intermediate files."""
   import igneous.tasks.image.ccl
-  src = cloudfiles.paths.normalize(src)
   igneous.tasks.image.ccl.clean_intermediate_files(src, mip)
 
 @cclgroup.command("auto")
-@click.argument("src")
-@click.argument("dest")
+@click.argument("src", type=CloudPath())
+@click.argument("dest", type=CloudPath())
 @click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
 @click.option('--mip', default=0, help="Apply to this level of the image pyramid.", show_default=True)
 @click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of destination layer. e.g. 128,128,64")
@@ -553,8 +547,6 @@ def ccl_auto(
   """
   For local volumes, execute all steps automatically.
   """
-  src = cloudfiles.paths.normalize(src)
-  dest = cloudfiles.paths.normalize(dest)
   parallel = int(ctx.obj.get("parallel", 1))
   tq = TaskQueue(normalize_path(queue))
   args = (queue, None, LEASE_SECONDS, True, -1, True, False)
@@ -646,8 +638,40 @@ def execute_helper(
 ):
   tq = TaskQueue(normalize_path(queue), region_name=aws_region)
 
+  sqs_sec_to_wait = 120
+  start_time = time.time()
+  last_empty = False
+
+  # Offical Amazon docs state that to determine if a queue is empty,
+  # you have to test whether it's consistently empty for several 
+  # minutes. So we pick 120 seconds to wait somewhat arbitrarily.
+  # https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/confirm-queue-is-empty.html
+  def is_empty():
+    nonlocal start_time
+    nonlocal sqs_sec_to_wait
+    nonlocal last_empty
+
+    if tq.path.protocol == "sqs":
+      if tq.is_empty():
+        if last_empty:
+          elapsed_time = time.time() - start_time
+          if elapsed_time >= sqs_sec_to_wait:
+            return True
+          else:
+            return False
+        else:
+          print(f"Queue appearently empty. Waiting {sqs_sec_to_wait} sec. to confirm.")
+          start_time = time.time()
+          last_empty = True
+          return False
+      else:
+        last_empty = False
+        return False
+    else:
+      return tq.is_empty()
+
   def stop_after_elapsed_time(tries, elapsed_time):
-    if exit_on_empty and tq.is_empty():
+    if exit_on_empty and is_empty():
       return True
 
     if min_sec < 0:
@@ -678,8 +702,8 @@ def meshgroup():
   pass
 
 @meshgroup.command("xfer")
-@click.argument("src")
-@click.argument("dest")
+@click.argument("src", type=CloudPath())
+@click.argument("dest", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option("--sharded", is_flag=True, default=False, help="Generate shard fragments instead of outputing mesh fragments.", show_default=True)
 @click.option("--dir", "mesh_dir", type=str, default=None, help="Write meshes into this directory instead of the one indicated in the info file.")
@@ -694,8 +718,6 @@ def mesh_xfer(
   Enables conversion of unsharded to sharded
   as well.
   """
-  src = cloudfiles.paths.normalize(src)
-  dest = cloudfiles.paths.normalize(dest)
   cv_src = CloudVolume(src)
 
   if not cv_src.mesh.meta.is_sharded() and sharded:
@@ -716,7 +738,7 @@ def mesh_xfer(
   tq.insert(tasks, parallel=parallel)
 
 @meshgroup.command("forge")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--mip', default=0, help="Perform meshing using this level of the image pyramid.", show_default=True)
 @click.option('--shape', type=Tuple3(), default=(448, 448, 448), help="Set the task shape in voxels.", show_default=True)
@@ -751,7 +773,6 @@ def mesh_forge(
 
   Sharded format not currently supports. Coming soon.
   """
-  path = cloudfiles.paths.normalize(path)
   if compress.lower() == "none":
     compress = False
 
@@ -769,7 +790,7 @@ def mesh_forge(
   tq.insert(tasks, parallel=parallel)
 
 @meshgroup.command("merge")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)")
 @click.option('--nlod', default=0, help="(multires) How many extra levels of detail to create.", show_default=True)
@@ -786,8 +807,6 @@ def mesh_merge(ctx, path, queue, magnitude, nlod, vqb, dir, min_chunk_size):
   a list of fragment files and uploading a "mesh manifest"
   file that is an index for locating the fragments.
   """
-  path = cloudfiles.paths.normalize(path)
-
   if nlod > 0:
     tasks = tc.create_unsharded_multires_mesh_tasks(
       path, num_lod=nlod, 
@@ -805,7 +824,7 @@ def mesh_merge(ctx, path, queue, magnitude, nlod, vqb, dir, min_chunk_size):
   tq.insert(tasks, parallel=parallel)
 
 @meshgroup.command("merge-sharded")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--nlod', default=1, help="Number of levels of detail to create.", type=int, show_default=True)
 @click.option('--vqb', default=16, help="Vertex quantization bits. Can be 10 or 16.", type=int, show_default=True)
@@ -813,6 +832,7 @@ def mesh_merge(ctx, path, queue, magnitude, nlod, vqb, dir, min_chunk_size):
 @click.option('--shard-index-bytes', default=2**13, help="Size in bytes to make the shard index.", type=int, show_default=True)
 @click.option('--minishard-index-bytes', default=2**15, help="Size in bytes to make the minishard index.", type=int, show_default=True)
 @click.option('--min-shards', default=1, help="Minimum number of shards to generate. Excess shards can make it easier to parallelize the merge process.", type=int, show_default=True)
+@click.option('--max-labels-per-shard', default=1000, help="Maximum average number of labels per a shard.", type=int, show_default=True)
 @click.option('--minishard-index-encoding', default="gzip", help="Minishard indices can be compressed. gzip or raw.", show_default=True)
 @click.option('--spatial-index-db', default=None, help="CloudVolume generated SQL database for spatial index.", show_default=True)
 @click.option('--min-chunk-size', type=Tuple3(), default="512,512,512",  help="(multires) Sets the minimum chunk size of the highest resolution mesh fragment.", show_default=True)
@@ -821,7 +841,8 @@ def mesh_sharded_merge(
   ctx, path, queue, 
   nlod, vqb, compress_level,
   shard_index_bytes, minishard_index_bytes, min_shards,
-  minishard_index_encoding, spatial_index_db, min_chunk_size
+  max_labels_per_shard, minishard_index_encoding, 
+  spatial_index_db, min_chunk_size
 ):
   """
   (2) Postprocess fragments into finished sharded multires meshes.
@@ -835,7 +856,6 @@ def mesh_sharded_merge(
   The shard and minishard index default sizes are set to
   accomodate efficient access for a 100 Mbps connection.
   """
-  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_sharded_multires_mesh_tasks(
     path, 
     num_lod=nlod,
@@ -847,6 +867,7 @@ def mesh_sharded_merge(
     min_shards=min_shards,
     spatial_index_db=spatial_index_db,
     min_chunk_size=min_chunk_size,
+    max_labels_per_shard=max_labels_per_shard,
   )
 
   parallel = int(ctx.obj.get("parallel", 1))
@@ -854,7 +875,7 @@ def mesh_sharded_merge(
   tq.insert(tasks, parallel=parallel)
 
 @meshgroup.command("rm")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)")
 @click.option('--dir', 'mesh_dir', default=None, help="Target this directory instead of the one indicated in the info file.")
@@ -863,7 +884,6 @@ def mesh_rm(ctx, path, queue, magnitude, mesh_dir):
   """
   Delete mesh files.
   """
-  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_mesh_deletion_tasks(
     path, magnitude=magnitude, mesh_dir=mesh_dir
   )
@@ -880,7 +900,7 @@ def spatialindexgroup():
   pass
 
 @spatialindexgroup.command("create")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--shape', default="448,448,448", type=Tuple3(), help="Shape in voxels of each indexing task.", show_default=True)
 @click.option('--mip', default=0, help="Perform indexing using this level of the image pyramid.", show_default=True)
@@ -895,7 +915,6 @@ def mesh_spatial_index_create(ctx, path, queue, shape, mip, fill_missing):
   This function provides a more efficient
   way to accomplish that than remeshing.
   """
-  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_spatial_index_mesh_tasks(
     cloudpath=path,
     shape=shape,
@@ -908,19 +927,26 @@ def mesh_spatial_index_create(ctx, path, queue, shape, mip, fill_missing):
   tq.insert(tasks, parallel=parallel)
 
 @spatialindexgroup.command("db")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.argument("database")
 @click.option('--progress', is_flag=True, default=False, help="Show progress bars.", show_default=True)
-def mesh_spatial_index_download(path, database, progress):
+@click.option('--allow-missing', is_flag=True, default=False, help="Allow missing index files.", show_default=True)
+@click.pass_context
+def mesh_spatial_index_download(ctx, path, database, progress, allow_missing):
   """
   Download the mesh spatial index into a database.
 
   sqlite paths: sqlite://filename.db (prefix optional)
   mysql paths: mysql://{user}:{pwd}@{host}/{database}
   """
-  path = cloudfiles.paths.normalize(path)
+  parallel = int(ctx.obj.get("parallel", 1))
   cv = CloudVolume(path)
-  cv.mesh.spatial_index.to_sql(database, progress=progress)
+  cv.mesh.spatial_index.to_sql(
+    database, 
+    progress=progress,
+    allow_missing=allow_missing, 
+    parallel=parallel,
+  )
 
 @main.group("skeleton")
 def skeletongroup():
@@ -936,7 +962,7 @@ def skeletongroup():
   pass
 
 @skeletongroup.command("forge")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--mip', default=0, help="Perform skeletonizing using this level of the image pyramid.", show_default=True)
 @click.option('--shape', type=Tuple3(), default=(512, 512, 512), help="Set the task shape in voxels.", show_default=True)
@@ -983,7 +1009,6 @@ def skeleton_forge(
 
   - https://github.com/seung-lab/kimimaro/wiki/The-Economics:-Skeletons-for-the-People
   """
-  path = cloudfiles.paths.normalize(path)
   teasar_params = {
     'scale': scale,
     'const': const, # physical units
@@ -1012,7 +1037,7 @@ def skeleton_forge(
 
 
 @skeletongroup.command("merge")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--min-cable-length', default=1000, help="Skip objects smaller than this physical path length. Default: 1000 nm", type=float)
 @click.option('--max-cable-length', default=None, help="Skip objects larger than this physical path length. Default: no limit", type=float)
@@ -1029,7 +1054,6 @@ def skeleton_merge(
   """
   (2) Postprocess fragments into finished skeletons.
   """
-  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_unsharded_skeleton_merge_tasks(
     path, 
     magnitude=magnitude, 
@@ -1044,7 +1068,7 @@ def skeleton_merge(
   tq.insert(tasks, parallel=parallel)
 
 @skeletongroup.command("merge-sharded")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--min-cable-length', default=1000, help="Skip objects smaller than this physical path length.", type=float, show_default=True)
 @click.option('--max-cable-length', default=None, help="Skip objects larger than this physical path length. Default: no limit", type=float)
@@ -1076,7 +1100,6 @@ def skeleton_sharded_merge(
   The shard and minishard index default sizes are set to
   accomodate efficient access for a 100 Mbps connection.
   """
-  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_sharded_skeleton_merge_tasks(
     path, 
     dust_threshold=min_cable_length,
@@ -1095,7 +1118,7 @@ def skeleton_sharded_merge(
   tq.insert(tasks, parallel=parallel)
 
 @imagegroup.command("rm")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--mip', default=0, help="Which mip level to start deleting from. Default: 0")
 @click.option('--num-mips', default=5, help="The number of mip levels to delete at once. Default: 5")
@@ -1108,7 +1131,6 @@ def delete_images(
   """
   Delete the image layer of a dataset.
   """
-  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_deletion_tasks(
     path, mip, num_mips=num_mips, shape=shape
   )
@@ -1117,7 +1139,7 @@ def delete_images(
   tq.insert(tasks, parallel=parallel)
 
 @skeletongroup.command("rm")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)")
 @click.option('--dir', 'skel_dir', default=None, help="Target this directory instead of the one indicated in the info file.")
@@ -1126,7 +1148,6 @@ def skeleton_rm(ctx, path, queue, magnitude, skel_dir):
   """
   Delete skeleton files.
   """
-  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_skeleton_deletion_tasks(
     path, magnitude=magnitude, skel_dir=skel_dir
   )
@@ -1137,8 +1158,8 @@ def skeleton_rm(ctx, path, queue, magnitude, skel_dir):
 
 
 @skeletongroup.command("xfer")
-@click.argument("src")
-@click.argument("dest")
+@click.argument("src", type=CloudPath())
+@click.argument("dest", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option("--sharded", is_flag=True, default=False, help="Generate shard fragments instead of outputing mesh fragments.", show_default=True)
 @click.option("--dir", "skel_dir", type=str, default=None, help="Write skeletons into this directory instead of the one indicated in the info file.")
@@ -1153,8 +1174,6 @@ def skel_xfer(
   Enables conversion of unsharded to sharded
   as well.
   """
-  src = cloudfiles.paths.normalize(src)
-  dest = cloudfiles.paths.normalize(dest)
   cv_src = CloudVolume(src)
 
   if not cv_src.skeleton.meta.is_sharded() and sharded:
@@ -1180,7 +1199,7 @@ def spatialindexgroupskel():
   pass
 
 @spatialindexgroupskel.command("create")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
 @click.option('--shape', default="512,512,512", type=Tuple3(), help="Shape in voxels of each indexing task.", show_default=True)
 @click.option('--mip', default=0, help="Perform indexing using this level of the image pyramid.", show_default=True)
@@ -1195,7 +1214,6 @@ def skel_spatial_index_create(ctx, path, queue, shape, mip, fill_missing):
   This function provides a more efficient
   way to accomplish that than remeshing.
   """
-  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_spatial_index_skeleton_tasks(
     cloudpath=path,
     shape=shape,
@@ -1208,19 +1226,26 @@ def skel_spatial_index_create(ctx, path, queue, shape, mip, fill_missing):
   tq.insert(tasks, parallel=parallel)
 
 @spatialindexgroupskel.command("db")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.argument("database")
 @click.option('--progress', is_flag=True, default=False, help="Show progress bars.", show_default=True)
-def skel_spatial_index_download(path, database, progress):
+@click.option('--allow-missing', is_flag=True, default=False, help="Allow missing index files.", show_default=True)
+@click.pass_context
+def skel_spatial_index_download(path, database, progress, allow_missing):
   """
   Download the skeleton spatial index into a database.
 
   sqlite paths: sqlite://filename.db (prefix optional)
   mysql paths: mysql://{user}:{pwd}@{host}/{database}
   """
-  path = cloudfiles.paths.normalize(path)
+  parallel = int(ctx.obj.get("parallel", 1))
   cv = CloudVolume(path)
-  cv.skeleton.spatial_index.to_sql(database, progress=progress)
+  cv.skeleton.spatial_index.to_sql(
+    database, 
+    progress=progress,
+    allow_missing=allow_missing, 
+    parallel=parallel,
+  )
 
 @main.group("design")
 def designgroup():
@@ -1230,7 +1255,7 @@ def designgroup():
   pass
 
 @designgroup.command("bounds")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
 def dsbounds(path, mip):
   """
@@ -1238,7 +1263,6 @@ def dsbounds(path, mip):
   an unsharded image volume. Useful when there
   is a corrupted info file.
   """
-  path = cloudfiles.paths.normalize(path)
   cv = CloudVolume(path, mip=mip)
   cf = CloudFiles(path)
 
@@ -1254,7 +1278,7 @@ def dsbounds(path, mip):
   print(f"Chunks: {chunk_size}")
 
 @designgroup.command("ds-memory")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.argument("memory_bytes", type=float)
 @click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
 @click.option('--factor', default="2,2,1", type=Tuple3(), help="Downsample factor to use.", show_default=True)
@@ -1264,8 +1288,7 @@ def dsmemory(path, memory_bytes, mip, factor, verbose, max_mips):
   """
   Compute the task shape that maximizes the number of
   downsamples for a given amount of memory.
-  """ 
-  path = cloudfiles.paths.normalize(path)
+  """
   cv = CloudVolume(path, mip=mip)
 
   data_width = np.dtype(cv.dtype).itemsize
@@ -1304,7 +1327,7 @@ def dsmemory(path, memory_bytes, mip, factor, verbose, max_mips):
     print(shape)
 
 @designgroup.command("ds-shape")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.argument("shape", type=Tuple3())
 @click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
 @click.option('--factor', default="2,2,1", type=Tuple3(), help="Downsample factor to use.", show_default=True)
@@ -1312,8 +1335,7 @@ def dsshape(path, shape, mip, factor):
   """
   Compute the approximate memory usage for a
   given downsample task shape.
-  """ 
-  path = cloudfiles.paths.normalize(path)
+  """
   cv = CloudVolume(path, mip=mip)
   data_width = np.dtype(cv.dtype).itemsize
   memory_bytes = memory_used(data_width, shape, factor)  
@@ -1334,7 +1356,7 @@ def memory_used(data_width, shape, factor):
   return memory_bytes
 
 @main.command("view")
-@click.argument("path")
+@click.argument("path", type=CloudPath())
 @click.option('--browser/--no-browser', default=True, is_flag=True, help="Open the dataset in the system's default web browser.")
 @click.option('--port', default=1337, help="localhost server port for the file server.", show_default=True)
 @click.option('--ng', default="https://neuroglancer-demo.appspot.com/", help="Alternative Neuroglancer webpage to use.", show_default=True)
@@ -1345,7 +1367,6 @@ def view(path, browser, port, ng):
   # later improvements: 
   #   could use local neuroglancer
   #   modify the url to autopopulate params to avoid a click
-  path = cloudfiles.paths.normalize(path)
   url = f"{ng}#!%7B%22layers%22:%5B%7B%22type%22:%22new%22%2C%22source%22:%22precomputed://http://localhost:{port}%22%2C%22tab%22:%22source%22%2C%22name%22:%22localhost:{port}%22%7D%5D%2C%22selectedLayer%22:%7B%22visible%22:true%2C%22layer%22:%22localhost:{port}%22%7D%2C%22layout%22:%224panel%22%7D"
   if browser:
     webbrowser.open(url, new=2)
